bcm_root <- Sys.getenv("BCM_ROOT")
source(paste(bcm_root, "scripts/plots_functions.r", sep="/"))

# Specify the model name here
model_name <- "shells2"

# The list of algorithm
folders <- list()
folders$Parallel_tempering <- "output_pt"
folders$Parallel_tempering_autoblock <- "output_pt_autoblock"
folders$Feedback_optimized_parallel_tempering <- "output_fopt"
folders$MultiNest <- "output_nested_multinest"
folders$Nested_ellipsoid <- "output_nested_ellipsoid"
folders$Nested_MCMC <- "output_nested_mcmc"
folders$SMC <- "output_smc"

# Load all the results
models <- list()
for (i in 1:length(folders)) {
  name <- names(folders)[i]
  cat("Loading", name, "\n")
  models[[name]] <- load_model(model_name, folders[[i]])
}
nvar <- models[[1]]$nvar

# Overview density plot
png(paste(model_name, "/density_overview.png", sep=""), width=1800, height=1000)
par(mar=c(2,2,3,2))
par(mfrow=c(nvar, length(folders)))
for (j in 1:nvar) {
  for (i in 1:length(folders)) {
    plot_variable_distribution(models[[i]], var_ix=j)
  }
}
res <- dev.off()

# Overview traces plot
png(paste(model_name, "/traces_overview.png", sep=""), width=1800, height=1000)
par(mar=c(2,2,3,2))
par(mfrow=c(nvar, length(folders)))
for (j in 1:nvar) {
  for (i in 1:length(folders)) {
    varattrs <- xmlAttrs(models[[i]]$vars[[j]])
    plot(models[[i]]$samples[,j], main=varattrs["name"], xlab="Sample index", ylab="Value")
  }
}
res <- dev.off()

# Compare various statistics between the samplers for a particular variable
variable_ix <- 2
sampler_comparison_overview <- data.frame(mean=NA, sd=NA, median=NA, q025=NA, q975=NA, autocorrelation_lag1=NA, decorrelation_lag=NA)
for (i in 1:length(models)) {
  s <- variable_summary(models[[i]])
  sampler_comparison_overview[i,] <- s[variable_ix,]
  rownames(sampler_comparison_overview)[i] <- names(models)[i]
}
sampler_comparison_overview
