#pragma once

#include "Prior.h"

namespace bcm {

class PriorVineCopula : public Prior
{
public:
	PriorVineCopula(VariableSet* varset);
	virtual ~PriorVineCopula();

	bool LoadFromXML(const boost::property_tree::ptree& xml_node);

	virtual bool EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const;
	virtual bool Sample(Real* values, RNG* rng) const;
	virtual bool EvaluateMarginalVariance(size_t i, Real& var) const;

protected:
	enum MarginalType
	{
		MT_Normal,
		MT_Beta,
		MT_Gamma,
		MT_ECDF_KDE,
		MT_Invalid
	};
	struct Marginal
	{
		Marginal() 
			: type(MT_Invalid)
			, min(std::numeric_limits<Real>::quiet_NaN())
			, max(std::numeric_limits<Real>::quiet_NaN())
		{}

		MarginalType type;
		VectorReal p;		// Proportions
		VectorReal param1;	// Normal: mu, Beta: a, Gamma: shape; ecdf: bw
		VectorReal param2;	// Normal: sigma, Beta: b, Gamma: scale; ecdf: x
		Real min;
		Real max;
		size_t copula_ix;
	};

	std::vector<Marginal> marginals;

	Eigen::MatrixXi M;
	Eigen::MatrixXi family;
	Eigen::MatrixXi MaxMat;
	Eigen::MatrixXi CoDirect;
	Eigen::MatrixXi CoIndirect;
	Eigen::MatrixXd par;
	Eigen::MatrixXd par2;

	Eigen::MatrixXi rM;
	Eigen::MatrixXi rfamily;
	Eigen::MatrixXi rMaxMat;
	Eigen::MatrixXi rCoDirect;
	Eigen::MatrixXi rCoIndirect;
	Eigen::MatrixXd rpar;
	Eigen::MatrixXd rpar2;
};

}
