#pragma once

#include "SMCSamplingStrategy.h"
#include "MVNRNG.h"

namespace bcm {

class SMCSamplingStrategyClusteredMCMC : public SMCSamplingStrategy
{
public:
	SMCSamplingStrategyClusteredMCMC(SamplerSMC* sampler, TaskManager* task_manager);
	virtual ~SMCSamplingStrategyClusteredMCMC();

	void SetStepsPerIteration(size_t n);
	void SetRecalculateSteps(size_t n);

protected:
	virtual bool UpdateStrategyIterationImpl(bool force_update, RNG* rng);
	virtual bool UpdateStrategyCurrentImpl(RNG* rng);
	virtual bool UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix);
	virtual bool CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix);

	void ClusterSamples(void* user, size_t thread);
	void CalculateBIC(size_t task);

	// Settings
	size_t recalculate_steps;
	size_t num_mcmc_steps_per_iteration;
	Real target_acceptance_rate;

	struct Cluster
	{
		VectorReal empirical_mean;
		VectorReal empirical_variance;
		MatrixReal empirical_covariance;
		MVNRNG mvn_sampler;
		VectorReal proposal_sigma;
		Real proposal_scale;
	};

	struct AsyncParams {
		VectorReal proposed_var_values;
		std::vector<size_t> num_proposed;
		std::vector<size_t> num_accepted;
	};
	std::vector<AsyncParams> async;

	size_t update_step;
	
	std::vector< std::vector<Cluster> > clusters;
	std::vector< std::vector<size_t> > cluster_assignments;
	std::vector< Real > clustering_bic;
	size_t use_k;

};

}
