#pragma once

#include "MVNRNG.h"

namespace bcm {

class SamplerPT;

class SamplerPTChain
{
public:
	SamplerPTChain(SamplerPT* sampler);
	~SamplerPTChain();

	bool Initialize(unsigned long rngseed);

	void SetTemperature(Real temperature);
	bool AdaptProposal();
	void SetMCMCMovesPerIteration(size_t num_moves);
	bool FindStartingPosition();
	bool MutateMove(bool burnin, size_t thread);
	void MutateMoveAsync(bool burnin);
	bool MutateMoveWait();
	bool ExchangeMove(bool burnin, SamplerPTChain& other);
	void LogStatistics();
	void LogProposalInfo();

	inline Real GetLogPrior() const { return lprior; }
	inline Real GetLogLikelihood() const { return llh; }
	inline Real GetLogPowerPosterior() const { return lpowerposterior; }
	inline const VectorReal& GetVarValues() const { return current_var_values; }
	inline const VectorReal& GetOutputValues() const { return current_output_values; }
	inline const VectorReal& GetEmpiricalVariance() const { return empirical_variance; }
	inline Real GetTemperature() const { return temperature; }

private:
	void AsyncDoMutateMove(void* user, size_t thread);
	void UpdateVariance();

	// Settings
	SamplerPT* sampler;
	Real temperature;
	size_t num_mcmc_moves_per_iteration;

	// Run-time variables
	RNG* rng;

	VectorReal current_var_values;
	VectorReal current_output_values;
	Real lprior;
	Real llh;
	Real lpowerposterior;

	size_t attempted_mutate;
	size_t attempted_exchange;
	size_t accepted_mutate;
	size_t accepted_exchange;

	// For blocked sampler
	struct Block
	{
		std::vector<size_t> variable_indices;
		MVNRNG mvn_sampler;	// For full covariance sampling
		VectorReal sigma;	// For diagonal covariance sampling
		Real scale;
		Real current_acceptance_rate_ema;
	};
	std::vector<Block> sampler_blocks;

	VectorReal empirical_mean;
	VectorReal empirical_mean_nm1;
	VectorReal empirical_variance;
	VectorReal empirical_variance_calc;
	MatrixReal empirical_covariance;
	MatrixReal empirical_correlation;
	MatrixReal empirical_correlation_calc;
	size_t empirical_n;
	size_t empirical_ix;

	// For parallelization
	uint64 async_task;
	bool async_burnin;
	volatile bool async_result;
};

}
