#pragma once

#include <sbml/SBMLTypes.h>
#include "SBMLRatelaws.h"
#include "VariableSet.h"

class SBMLAssignmentRule
{
public:
	SBMLAssignmentRule();
	~SBMLAssignmentRule();
	
	bool Initialize(const Rule* rule, const SBMLModel* model);
	bool SetSpeciesVector(const std::vector<std::string>& species_ids);
	bool SetVariableSet(const bcm::VariableSet* variables);
	bool PostInitialize();

	bool Calculate(const Real* species, const Real* parameters, Real* value) const;
	inline const std::string& GetVariable() const { return TargetVariable; }

private:
	void MapRateLawSpecies(const std::vector<std::string>& species_ids, const ASTNode* node);
	void MapRateLawParameters(const bcm::VariableSet* variables, const ASTNode* node);

	std::string TargetVariable;
	std::map<std::string, size_t> species_index_map;
	std::map<std::string, size_t> parameter_index_map;
	ASTNode* rate_law;
	SBMLRatelawElement* evaluate_rate_law;
};
