#include "Utils.h"
#include "CVODESolver.h"
#include "LinearSolverMKL.h"
#include "SBMLAssignmentRule.h"
#include "SBMLModel.h"
#include "SBMLModelParameters.h"
#include "SBMLReaction.h"
#include "SBMLSpecies.h"

#define USE_CODE 0
void derivative(double* out, const double* species, const double* parameters)
{
	out[0] = +(parameters[35]*species[17])-(parameters[0]*species[0]);
	out[1] = +((parameters[15]+(parameters[21]*species[0]))/(1+exp((parameters[30]*(log(species[4])-parameters[32])))))-(parameters[10]*species[1]);
	out[2] = +(parameters[19]+(parameters[27]*species[16]))-(parameters[10]*species[2]);
	out[3] = +(parameters[18]+(parameters[26]*species[6]))-(parameters[10]*species[3]);
	out[4] = +(parameters[35]*species[5])-(parameters[5]*species[4]);
	out[5] = +(parameters[16]+(parameters[22]*species[14]))-(parameters[10]*species[5]);
	out[6] = +(parameters[35]*species[7])-(parameters[6]*species[6]);
	out[7] = -(parameters[10]*species[7])+(parameters[17]+(parameters[23]*species[14]));
	out[8] = -(parameters[1]*species[8])+(parameters[35]*species[9]);
	out[9] = +(parameters[12]+(parameters[24]*species[14]))-(parameters[10]*species[9]);
	out[10] = +(parameters[35]*species[12])-(parameters[2]*species[10]);
	out[11] = +(parameters[35]*species[13])-(parameters[3]*species[11]);
	out[12] = +(parameters[13]+(parameters[25]*species[14]))-(parameters[10]*species[12]);
	out[13] = +(parameters[14]+(parameters[28]*species[16]))-(parameters[10]*species[13]);
	out[14] = +(parameters[35]*species[1])-(parameters[4]*species[14]);
	out[15] = +(parameters[35]*species[2])-(parameters[8]*species[15]);
	out[16] = +(parameters[35]*species[3])-((parameters[7]*species[16])+((parameters[9]*species[16])/(1+exp((parameters[33]*(log(species[8])-parameters[34]))))));
	out[17] = -(parameters[10]*species[17])+((parameters[11]+(parameters[20]*species[15]))/(1+exp((parameters[29]*(log(species[4])-parameters[31])))));
}

void jacobian(double** out, const double* species, const double* parameters)
{
	out[0][0] = -(parameters[0]*(OdeReal)1.0);
	out[0][1] = +((+(parameters[21]*(OdeReal)1.0))/(1+exp((parameters[30]*(log(species[4])-parameters[32])))));

	out[1][1] = -(parameters[10]*(OdeReal)1.0);
	out[1][14] = +(parameters[35]*(OdeReal)1.0);

	out[2][2] = -(parameters[10]*(OdeReal)1.0);
	out[2][15] = +(parameters[35]*(OdeReal)1.0);

	out[3][3] = -(parameters[10]*(OdeReal)1.0);
	out[3][16] = +(parameters[35]*(OdeReal)1.0);

	out[4][1] = +(-(+exp((parameters[30]*(log(species[4])-parameters[32])))*((parameters[30]*(OdeReal)1.0/(species[4]))))*(parameters[15]+(parameters[21]*species[0]))/((1+exp((parameters[30]*(log(species[4])-parameters[32]))))*(1+exp((parameters[30]*(log(species[4])-parameters[32]))))));
	out[4][4] = -(parameters[5]*(OdeReal)1.0);
	out[4][17] = +(-(+exp((parameters[29]*(log(species[4])-parameters[31])))*((parameters[29]*(OdeReal)1.0/(species[4]))))*(parameters[11]+(parameters[20]*species[15]))/((1+exp((parameters[29]*(log(species[4])-parameters[31]))))*(1+exp((parameters[29]*(log(species[4])-parameters[31]))))));

	out[5][4] = +(parameters[35]*(OdeReal)1.0);
	out[5][5] = -(parameters[10]*(OdeReal)1.0);

	out[6][3] = +(+(parameters[26]*(OdeReal)1.0));
	out[6][6] = -(parameters[6]*(OdeReal)1.0);

	out[7][6] = +(parameters[35]*(OdeReal)1.0);
	out[7][7] = -(parameters[10]*(OdeReal)1.0);

	out[8][8] = -(parameters[1]*(OdeReal)1.0);
	out[8][16] = -(+(-(+exp((parameters[33]*(log(species[8])-parameters[34])))*((parameters[33]*(OdeReal)1.0/(species[8]))))*(parameters[9]*species[16])/((1+exp((parameters[33]*(log(species[8])-parameters[34]))))*(1+exp((parameters[33]*(log(species[8])-parameters[34])))))));

	out[9][8] = +(parameters[35]*(OdeReal)1.0);
	out[9][9] = -(parameters[10]*(OdeReal)1.0);

	out[10][10] = -(parameters[2]*(OdeReal)1.0);

	out[11][11] = -(parameters[3]*(OdeReal)1.0);

	out[12][10] = +(parameters[35]*(OdeReal)1.0);
	out[12][12] = -(parameters[10]*(OdeReal)1.0);

	out[13][11] = +(parameters[35]*(OdeReal)1.0);
	out[13][13] = -(parameters[10]*(OdeReal)1.0);

	out[14][5] = +(+(parameters[22]*(OdeReal)1.0));
	out[14][7] = +(+(parameters[23]*(OdeReal)1.0));
	out[14][9] = +(+(parameters[24]*(OdeReal)1.0));
	out[14][12] = +(+(parameters[25]*(OdeReal)1.0));
	out[14][14] = -(parameters[4]*(OdeReal)1.0);

	out[15][15] = -(parameters[8]*(OdeReal)1.0);
	out[15][17] = +((+(parameters[20]*(OdeReal)1.0))/(1+exp((parameters[29]*(log(species[4])-parameters[31])))));

	out[16][2] = +(+(parameters[27]*(OdeReal)1.0));
	out[16][13] = +(+(parameters[28]*(OdeReal)1.0));
	out[16][16] = -(+(parameters[7]*(OdeReal)1.0)+((parameters[9]*(OdeReal)1.0)/(1+exp((parameters[33]*(log(species[8])-parameters[34]))))));

	out[17][0] = +(parameters[35]*(OdeReal)1.0);
	out[17][17] = -(parameters[10]*(OdeReal)1.0);

}

SBMLModel::SBMLModel(size_t numthreads)
	: model(NULL)
	, variables(NULL)
{
	Solvers.resize(numthreads);
}

SBMLModel::~SBMLModel()
{
	for (size_t i = 0; i < Solvers.size(); i++) {
		bcm::safe_delete(Solvers[i].solver);
		bcm::safe_delete(Solvers[i].linear_solver);
		bcm::safe_delete(Solvers[i].parameters);
	}
	for (std::map<std::string, SBMLSpecies*>::iterator si = Species.begin(); si != Species.end(); ++si) {
		bcm::safe_delete(si->second);
	}
	for (std::map<std::string, SBMLReaction*>::iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
		bcm::safe_delete(ri->second);
	}
	for (size_t i = 0; i < AssignmentRules.size(); i++) {
		bcm::safe_delete(AssignmentRules[i]);
	}
}

bool SBMLModel::LoadSBML(const std::string& fn)
{
	document = readSBML(fn.c_str());
	if (document->getNumErrors() > 0) {
		LOGERROR("Errors reading SBML file %s: %d", fn.c_str(), document->getNumErrors());
		for (unsigned int i = 0; i < document->getNumErrors(); i++) {
			LOGERROR(" %d: %s", i, document->getError(i)->getMessage().c_str());
		}
		bcm::safe_delete(document);
		return false;
	}

	model = document->getModel();

	for (unsigned int i = 0; i < model->getNumSpecies(); i++) {
		SBMLSpecies* sp = new SBMLSpecies;
		if (!sp->Initialize(model->getSpecies(i), model->getAnnotation())) {
			return false;
		}
		if (Species.find(sp->GetId()) != Species.end()) {
			LOGERROR("Duplicate species id %s", sp->GetId().c_str());
			return false;
		}
		Species[sp->GetId()] = sp;
	}

	for (unsigned int i = 0; i < model->getNumReactions(); i++) {
		SBMLReaction* re = new SBMLReaction;
		if (!re->Initialize(model->getReaction(i), this)) {
			return false;
		}
		if (Reactions.find(re->GetId()) != Reactions.end()) {
			LOGERROR("Duplicate reaction id %s", re->GetId().c_str());
			return false;
		}
		Reactions[re->GetId()] = re;
	}

	// Retrieve the species to simulate
	std::set<std::string> species_without_reactions;
	for (std::map<std::string, SBMLSpecies*>::iterator si = Species.begin(); si != Species.end(); ++si) {
		//if (si->second->GetType() != SBMLSpecies::Sink && si->second->GetType() != SBMLSpecies::Gene) {
		if (si->second->GetType() != SBMLSpecies::Sink) {
			SimulatedSpecies.push_back(si->first);
			species_without_reactions.insert(si->first);
		}
	}

	for (std::map<std::string, SBMLReaction*>::iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
		if (!ri->second->SetSpeciesVector(SimulatedSpecies)) {
			return false;
		}
		for (std::vector<std::string>::const_iterator rti = ri->second->GetReactants().begin(); rti != ri->second->GetReactants().end(); ++rti) {
			std::set<std::string>::iterator si = species_without_reactions.find(*rti);
			if (si != species_without_reactions.end()) {
				species_without_reactions.erase(si);
			}
		}
		for (std::vector<std::string>::const_iterator pdi = ri->second->GetProducts().begin(); pdi != ri->second->GetProducts().end(); ++pdi) {
			std::set<std::string>::iterator si = species_without_reactions.find(*pdi);
			if (si != species_without_reactions.end()) {
				species_without_reactions.erase(si);
			}
		}
	}
	
#if 0
	// Do not simulate species which are neither a reactant nor a product
	for (std::vector<std::string>::iterator ssi = SimulatedSpecies.begin(); ssi != SimulatedSpecies.end(); ) {
		std::set<std::string>::iterator si = species_without_reactions.find(*ssi);
		if (si != species_without_reactions.end()) {
			ssi = SimulatedSpecies.erase(ssi);
		} else {
			++ssi;
		}
	}
#endif

	// Initialize assignment rules
	SpeciesIsAssigned.resize(SimulatedSpecies.size(), false);
	for (unsigned int i = 0; i < model->getNumRules(); i++) {
		const Rule* rule = model->getRule(i);
		if (rule->isAssignment()) {
			SBMLAssignmentRule* newrule = new SBMLAssignmentRule;
			if (!newrule->Initialize(rule, this)) {
				return false;
			}
			newrule->SetSpeciesVector(SimulatedSpecies);
			AssignmentRules.push_back(newrule);

			const std::string& target = newrule->GetVariable();
			bool found = false;
			for (size_t ix = 0; ix < SimulatedSpecies.size(); ix++) {
				if (SimulatedSpecies[ix] == target) {
					AssignmentRulesTargetIx.push_back(ix);
					SpeciesIsAssigned[ix] = true;
					found = true;
					break;
				}
			}

			if (!found) {
				LOGERROR("Target variable \"%s\" for assignment rule %d not found!", target.c_str(), i);
				return false;
			}
		}
	}

	// TODO Check that there are no two species with the same name (if necessary?)
	
	for (size_t i = 0; i < Solvers.size(); i++) {
		Solvers[i].parameters = new SBMLModelParameters(this);
	}

	return true;
}

bool SBMLModel::SetVariableSet(const bcm::VariableSet* variables)
{
	this->variables = variables;
	for (size_t i = 0; i < Solvers.size(); i++) {
		if (!Solvers[i].parameters->SetVariableSet(variables)) {
			return false;
		}
	}
	return true;
}

bool SBMLModel::Initialize()
{
	if (!variables) {
		LOGERROR("VariableSet has not been set");
		return false;
	}

	// Map variables to parameters
	for (std::map<std::string, SBMLReaction*>::iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
		if (!ri->second->SetVariableSet(variables)) {
			return false;
		}
		if (!ri->second->PostInitialize()) {
			return false;
		}
	}
	for (size_t ri = 0; ri < AssignmentRules.size(); ri++) {
		if (!AssignmentRules[ri]->SetVariableSet(variables)) {
			return false;
		}
		if (!AssignmentRules[ri]->PostInitialize()) {
			return false;
		}
	}

	// Initialize CVODE solvers
	for (size_t i = 0; i < Solvers.size(); i++) {
		CVODESolver::TDeriviativeFunction derivative = boost::bind(&SBMLModel::CalculateDerivative, this, _1, _2, _3, _4);

		Solvers[i].solver = new CVODESolver();
		//Solvers[i].linear_solver = new LinearSolverMKL(this);
		//Solvers[i].solver->SetLinearSolver(Solvers[i].linear_solver);
		Solvers[i].solver->SetDerivativeFunction(derivative);
#if USE_CODE
		CVODESolver::TJacobianFunction jacobian = boost::bind(&SBMLModel::CalculateJacobian, this, _1, _2, _3, _4, _5);
		Solvers[i].solver->SetJacobianFunction(jacobian);
#endif
		if (!Solvers[i].solver->Initialize(SimulatedSpecies.size(), Solvers[i].solver)) {
			return false;
		}
		Solvers[i].y.resize(SimulatedSpecies.size());
		Solvers[i].transformed_variables = VectorReal::Zero(variables->GetNumVariables());
		Solvers[i].transformed_variables_use_in_ode = OdeVectorReal::Zero(variables->GetNumVariables());
		
#if 1
		VectorReal tolerance = VectorReal::Zero(SimulatedSpecies.size());
		for (size_t si = 0; si < SimulatedSpecies.size(); si++) {
			const SBMLSpecies* species = GetSpecies(SimulatedSpecies[si]);
			if (species->GetType() == SBMLSpecies::Transcript) {
				tolerance(si) = 1e-10;
			} else {
				tolerance(si) = 1e-6;
			}
		}
		Solvers[i].solver->SetTolerance(1e-5, tolerance);
#else
		Solvers[i].solver->SetTolerance(1e-5, 1e-5);
#endif
	}

	return true;
}

void SBMLModel::DumpODEStatistics(const std::string& base_filename)
{
	for (size_t i = 0; i < Solvers.size(); i++) {
		std::string fn = base_filename + std::to_string((uint64)i) + std::string(".tsv");
		Solvers[i].solver->DumpStatistics(fn.c_str());
	}
}

std::string SBMLModel::GenerateCode()
{
	std::string code;

	code += "void derivative(double* out, const double* species, const double* parameters)\n";
	code += "{\n";

	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		std::string eqn;
		for (std::map<std::string, SBMLReaction*>::const_iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
			ri->second->AddRateEquation(i, eqn);
		}
		if (eqn.empty()) {
			code += "\tout[" + std::to_string((uint64)i) + "] = 0.0;\n";
		} else {
			code += "\tout[" + std::to_string((uint64)i) + "] = " + eqn + ";\n";
		}
	}

	code += "}\n\n";

	code += "void jacobian(double** out, const double* species, const double* parameters)\n";
	code += "{\n";

	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		for (size_t j = 0; j < SimulatedSpecies.size(); j++) {
			std::string eqn;
			for (std::map<std::string, SBMLReaction*>::const_iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
				ri->second->AddRateDerivativeEquation(j, i, eqn);
			}
#if 1
			if (!eqn.empty()) {
				code += "\tout[" + std::to_string((uint64)i) + "][" + std::to_string((uint64)j) + "] = " + eqn + ";\n";
			}
#else
			if (!eqn.empty()) {
				code += "\tM.coeffRef(" + std::to_string((uint64)j) + "," + std::to_string((uint64)i) + ") = " + eqn + "; // j=" + std::to_string((uint64)i) + "\n";
			}
#endif
		}
		code += "\n";
	}

	code += "}\n";

	return code;
}

std::string SBMLModel::GenerateJacobianCode(size_t i, size_t j)
{
	std::string eqn;
	for (std::map<std::string, SBMLReaction*>::const_iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
		ri->second->AddRateDerivativeEquation(j, i, eqn);
	}
	return eqn;
}

void SBMLModel::ResetForcedVariables(size_t threadix)
{
	ParallelSolver& s = Solvers[threadix];
	s.parameters->ResetForcedVariables();
}

bool SBMLModel::SetForcedVariable(const std::string& variable_name, Real value, size_t threadix)
{
	ParallelSolver& s = Solvers[threadix];
	return s.parameters->SetForcedVariable(variable_name, value);
}

bool SBMLModel::Simulate(const Real* variable_values, const std::vector<Real>& timepoints, size_t threadix)
{
	ParallelSolver& s = Solvers[threadix];

	// Store variables pointer for use in derivative evaluation
	s.variables = variable_values;
	for (size_t i = 0; i < variables->GetNumVariables(); i++) {
		if (variables->IsVariableLogSpace(i)) {
			s.transformed_variables(i) = pow(10.0, variable_values[i]);
		} else {
			s.transformed_variables(i) = variable_values[i];
		}
	}

	// Ensure output is big enough
	if (s.trajectories.cols() != timepoints.size()) {
		s.trajectories.setZero(SimulatedSpecies.size(), timepoints.size());
	}

	// Force parameters to specific value if mandated by an experimental condition
	s.parameters->UpdateVariableValues(s.transformed_variables.data());

	// Retrieve initial conditions
	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		if (SpeciesIsAssigned[i]) {
			// Will be set a few lines below by CalculateAssignments
		} else {
			const SBMLSpecies* sp = GetSpecies(SimulatedSpecies[i]);
			std::string param_name = "initial_value_" + sp->GetFullName();
			size_t ix = variables->GetVariableIndex(param_name);
			if (ix == std::numeric_limits<size_t>::max()) {
				s.y[i] = std::numeric_limits<Real>::quiet_NaN();
			} else {
				s.y[i] = s.transformed_variables(ix);
			}
		}
	}
	s.parameters->UpdateInitialValues(s.y.data());
	CalculateAssignments(s.y.data(), s.transformed_variables.data(), s.y.data());
	
	// Sanity check
	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		if (!(s.y[i] == s.y[i])) {
			LOGERROR("Invalid initial value for species %d/\"%s\"/\"%s\"", i, SimulatedSpecies[i].c_str(), GetSimulatedSpeciesFullName(i).c_str());
			LOGERROR("Has an initial value been specified in the prior or as an experimental condition?");
			return false;
		}
	}

	for (size_t i = 0; i < variables->GetNumVariables(); i++) {
		s.transformed_variables_use_in_ode(i) = (OdeReal)s.transformed_variables(i);
	}
	if (s.linear_solver) {
		s.linear_solver->SetParameters(s.transformed_variables_use_in_ode.data());
	}

	bool result = s.solver->Simulate(&s.y[0], timepoints, s.trajectories);
	if (!result) {
#if 0
		LOGERROR("Model simulation  failed with variable values:");
		for (size_t vi = 0; vi < variables->GetNumVariables(); vi++) {
			LOGERROR("  %s - %g", variables->GetVariableName(vi).c_str(), variable_values[vi]);
		}
#endif
		return false;
	}

	// Calculate assignments - NOTE that these assignments are not in effect during the simulation..
	for (size_t ti = 0; ti < timepoints.size(); ti++) {
		VectorReal timepoint = s.trajectories.col(ti);
		CalculateAssignments(timepoint.data(), s.transformed_variables.data(), timepoint.data());
		s.trajectories.col(ti) = timepoint;
	}

	return true;
}

bool SBMLModel::GetOutput(size_t species_ix, VectorReal& output, size_t threadix)
{
	if (species_ix >= SimulatedSpecies.size()) {
		LOGERROR("Invalid species index");
		return false;
	}

	ParallelSolver& s = Solvers[threadix];
	output = s.trajectories.row(species_ix);
	return true;
}

bool SBMLModel::GetOutput(const std::string& species_full_name, VectorReal& output, size_t threadix)
{
	ParallelSolver& s = Solvers[threadix];

	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		if (Species[SimulatedSpecies[i]]->GetFullName() == species_full_name) {
			output = s.trajectories.row(i);
			return true;
		}
	}

	LOGERROR("Could not find species with full name %s", species_full_name.c_str());
	return false;
}

void SBMLModel::GetParameters(std::vector<std::string>& parameters) const
{
	parameters.resize(model->getNumParameters());
	for (unsigned int i = 0; i < model->getNumParameters(); i++) {
		parameters[i] = model->getParameter(i)->getId();
	}
}

size_t SBMLModel::GetNumSimulatedSpecies() const
{
	return SimulatedSpecies.size();
}

const std::string& SBMLModel::GetSimulatedSpeciesName(size_t i) const
{
	return SimulatedSpecies[i];
}

std::string SBMLModel::GetSimulatedSpeciesFullName(size_t i) const
{
	std::map<std::string, SBMLSpecies*>::const_iterator iter = Species.find(SimulatedSpecies[i]);
	return iter->second->GetFullName();
}

void SBMLModel::GetSimulatedSpeciesFullNames(std::vector<std::string>& species_names) const
{
	species_names.clear();
	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		std::map<std::string, SBMLSpecies*>::const_iterator iter = Species.find(SimulatedSpecies[i]);
		species_names.push_back(iter->second->GetFullName());
	}
}

const SBMLSpecies* SBMLModel::GetSpecies(const std::string& id) const
{
	std::map<std::string, SBMLSpecies*>::const_iterator iter = Species.find(id);
	if (iter != Species.end()) {
		return iter->second;
	} else {
		return NULL;
	}
}

const SBMLSpecies* SBMLModel::GetSpeciesFromFullName(const std::string& id) const
{
	for (std::map<std::string, SBMLSpecies*>::const_iterator si = Species.begin(); si != Species.end(); ++si) {
		if (si->second->GetFullName() == id) {
			return si->second;
		}
	}
	return NULL;
}


bool SBMLModel::CalculateDerivative(OdeReal t, const OdeReal* y, OdeReal* dydt, void* user) const
{
	// Find which parallel solver was used
	size_t threadix;
	for (size_t i = 0; i < Solvers.size(); i++) {
		if (Solvers[i].solver == user) {
			threadix = i;
			break;
		}
	}

#if USE_CODE
	derivative(dydt, y, Solvers[threadix].transformed_variables_use_in_ode.data());
#else
	for (size_t i = 0; i < SimulatedSpecies.size(); i++) {
		dydt[i] = 0.0;
	}
	for (std::map<std::string, SBMLReaction*>::const_iterator ri = Reactions.begin(); ri != Reactions.end(); ++ri) {
		ri->second->AddRates(y, Solvers[threadix].transformed_variables_use_in_ode.data(), dydt);
	}
#endif

	return true;
}

bool SBMLModel::CalculateJacobian(OdeReal t, const OdeReal* y, const OdeReal* dydt, OdeReal** jac, void* user) const
{
	// Find which parallel solver was used
	size_t threadix;
	for (size_t i = 0; i < Solvers.size(); i++) {
		if (Solvers[i].solver == user) {
			threadix = i;
			break;
		}
	}
	
#if USE_CODE
	jacobian(jac, y, Solvers[threadix].transformed_variables_use_in_ode.data());
#endif

	return true;
}

void SBMLModel::CalculateAssignments(const Real* y, const Real* variables, Real* out) const
{
	for (size_t ri = 0; ri < AssignmentRules.size(); ri++) {
		Real value;
		AssignmentRules[ri]->Calculate(y, variables, &value);
		out[AssignmentRulesTargetIx[ri]] = value;
	}
}
