#pragma once

#include <sbml/SBMLTypes.h>
#include "VariableSet.h"

class SBMLModel;
class SBMLSpecies;
class SBMLRatelawElement;

class SBMLReaction
{
public:
	SBMLReaction();
	~SBMLReaction();

	enum EType {
		Association,
		Dissociation,
		Transition,
		CatalyzedTransition,
		Transcription,
		Translation,
		PhysicalStimulation,

		Invalid
	};

	bool Initialize(const Reaction* reaction, const SBMLModel* model);
	bool SetSpeciesVector(const std::vector<std::string>& species_ids);
	bool SetVariableSet(const bcm::VariableSet* variables);
	bool AddRates(const Real* species, const Real* parameters, Real* rates);
	void AddRateEquation(size_t species_ix, std::string& eqn);
	void AddRateDerivativeEquation(size_t species_ix, size_t deriv_ix, std::string& eqn);
	bool PostInitialize();

	inline const std::string& GetId() const { return id; }
	inline const std::vector<std::string>& GetReactants() const { return reactants; }
	inline const std::vector<std::string>& GetProducts() const { return products; }

private:
	void MapRateLawSpecies(const std::vector<std::string>& species_ids, const ASTNode* node);
	void MapRateLawParameters(const bcm::VariableSet* variables, const ASTNode* node);

	std::string id;
	EType type;
	std::vector<std::string> reactants;
	std::vector<std::string> products;
	std::vector<size_t> reactant_indices;
	std::vector<size_t> product_indices;
	std::map<std::string, size_t> species_index_map;
	std::map<std::string, size_t> parameter_index_map;

	ASTNode* rate_law;
	SBMLRatelawElement* evaluate_rate_law;
};
