#pragma once

#include "NetCDFDataFile.h"
#include "VariableSet.h"

class SBMLModel;

class LikelihoodPart
{
public:
	LikelihoodPart();
	virtual ~LikelihoodPart();

	virtual bool Load(const boost::property_tree::ptree& xml_node, const bcm::NetCDFDataFile* datafile);
	virtual bool SetVariableSet(const bcm::VariableSet* variables);
	virtual bool GetRequiredTimepoints(std::vector<Real>& timepoints);
	virtual bool EvaluateLogProbability(size_t threadix, const Real* values, SBMLModel* model, std::vector<Real> simulated_time, Real& logp) = 0;

protected:
	bool ParseDataName(const bcm::NetCDFDataFile* datafile);
	bool LookupDataValue1D(const bcm::NetCDFDataFile* datafile, size_t ix, Real* value);
	bool LookupDataValue2D(const bcm::NetCDFDataFile* datafile, size_t ix1, size_t ix2, Real* value);
	void PreEvaluateLogPdf(const Real* values, Real& optim);
	Real EvaluateLogPdf(const Real* values, Real modeled, Real observed, Real* optim = NULL);

	enum ELikelihoodFunction
	{
		LF_Normal,
		LF_StudentsT,
		LF_Delta,
		LF_Invalid
	};

	std::string SpeciesName;
	std::string DataGroup;
	std::string DataName;
	std::string DataNameWithoutIndex;
	std::vector<size_t> DataIndices;

	ELikelihoodFunction LikelihoodFunction;
	std::string VarianceParameterName;
	size_t VarianceParameterIx;
	bool VarianceParameterLog;
	Real T_Nu;
	std::string ScaleParameterName;
	size_t ScaleParameterIx;
};
