#pragma once

#include "NetCDFDataFile.h"
#include "VariableSet.h"

class SignalingNetwork;

class SBMLPDExperiment
{
public:
	virtual ~SBMLPDExperiment();
	
	virtual bool SetupAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_count) = 0;
	virtual void HandleAdditionalOutput(bcm::NetCDFDataFile* file, size_t sample_ix, const Real* values, size_t& value_ix) = 0;
	virtual bool EvaluateLogProbability(size_t threadix, const Real* values, Real& logp, std::vector<SBMLPDExperiment*> other_experiments) = 0;
	virtual bool GetOutput(size_t threadix, Real* output_values, size_t& numvalues) = 0;

	static SBMLPDExperiment* Create(const boost::property_tree::ptree& xml_node, SignalingNetwork* network, const bcm::VariableSet* varset, std::vector<SBMLPDExperiment*> other_experiments, size_t numthreads);
	inline const std::string& GetName() const { return Name; }

protected:
	struct ModelConditions
	{
		std::string model_name;
		VectorReal values;
		size_t model_ix;
	};

	struct ExpressionLevel
	{
		std::string model_name;
		VectorReal values;
		size_t model_ix;
		size_t parameter_ix;
	};

	struct DataPartBase
	{
		DataPartBase();

		std::string model_name;
		std::string data_name;
		std::string base_scale_var_suffix;

		bool use_base;
		bool use_scale;
		bool fixed_var;
		Real weight;
		
		size_t model_ix;
		size_t parameter_base_ix;
		size_t parameter_scale_ix;
		size_t parameter_variance_ix;
		Real fixed_variance_value;
		bool data_is_inactive_form;

		size_t model_ix2;
		size_t data_sum_parameter_ix;

		std::string relative_to_experiment;
	};
	
	struct ParallelDataBase {
		std::vector<Real> transformed_values;
		std::vector<Real> data_logC;
		VectorReal expression;
		std::vector<VectorReal> relative_to_experiment_values;
	};
	
	SBMLPDExperiment(size_t numthreads);
	bool Load(const boost::property_tree::ptree& xml_node, std::vector<SBMLPDExperiment*> other_experiments);
	bool ParseDataPartBase(const boost::property_tree::ptree& xml_node, DataPartBase& p, std::vector<SBMLPDExperiment*> other_experiments);
	bool ParseDataFileReference(const std::string& data_name, std::string& data_name_without_index, std::vector<size_t>& data_indices, const bcm::NetCDFDataFile* datafile);
	void PrepareActivitiesCalculation(VectorReal& activities, VectorReal& expression, const Real* transformed_values, size_t cell_ix);

	virtual bool LoadTypeSpecificNodes(const boost::property_tree::ptree& xml_node, const bcm::NetCDFDataFile* datafile) { return true; }
	virtual bool InitializeParallelData() = 0;
	virtual bool ParseDataNode(const boost::property_tree::ptree& xml_node, std::vector<SBMLPDExperiment*> other_experiments, const bcm::NetCDFDataFile* datafile) = 0;

	size_t numthreads;
	SignalingNetwork* network;
	const bcm::VariableSet* varset;

	std::string Name;
	std::string DataFile;

	std::vector<std::string> cell_lines;
	std::vector<ExpressionLevel> expression_levels;
	std::vector<ModelConditions> conditions;
};
