#pragma once

#include <boost/date_time/posix_time/posix_time.hpp>

#define LOG(message, ...)			do { bcm::logger->Log(bcm::Logger::Info   , __FILE__, __LINE__, message, ## __VA_ARGS__); } while (0)
#define LOGWARNING(message, ...)	do { bcm::logger->Log(bcm::Logger::Warning, __FILE__, __LINE__, message, ## __VA_ARGS__); } while (0)
#define LOGERROR(message, ...)		do { bcm::logger->Log(bcm::Logger::Error  , __FILE__, __LINE__, message, ## __VA_ARGS__); } while (0)

namespace bcm {

class Logger
{
public:
	enum ELevel
	{
		Info	= 0,
		Warning = 1,
		Error	= 2,
		None	= 3,

		Count
	};

	Logger();
	~Logger();

	bool SetLogToConsole(ELevel level);
	bool SetLogToFile(ELevel level, const char* filename);
	bool SetLogToBuffer(ELevel level, size_t size = 512 * 1024);

	inline const char* GetBuffer() const { return Buffer; }
	inline size_t GetBufferSize() const { return (Buffer ? BufferSize : 0); }

	void Log(ELevel level, const char* source_file, int line, const char* message, ...);
	void LogTime();
	void LogCPUInfo();

private:
	void LogConsole(ELevel level, const char* message);
	void LogFile(ELevel level, const char* message);
	void LogBuffer(ELevel level, const char* message);

	FILE* File;
	char* Buffer;
	size_t BufferOffset;
	size_t BufferSize;
	char* MessageBuffer;

	ELevel ConsoleLevel;
	ELevel FileLevel;
	ELevel BufferLevel;

	boost::posix_time::ptime LogStartTime;

	boost::mutex LogMutex;
};

extern Logger* logger;

}
